"use strict";

const PUPPETEER = require("puppeteer");
const FS = require("fs");
async function SLEEP(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
}

class WTRM {
    constructor() {
        this.config = {
            "page-width": 1920,
            "page-height": 1080,
            "timeout": 60000,
            "screenshots": false,
            "pretty": false,
            "dump-html-on-failed-step": false,
            "workflow": false,
            "user-data-dir": false,
            "html-file": false,
            "no-sandbox": false,
            "host-id": 0,
            "service-id": 0,
            "log-file": "/var/log/bloonix/bloonix-wtrm-js.log",
            "allow-download": false,
        };

        const argv = process.argv.slice();
        argv.shift(0); // node
        argv.shift(0); // bloonix-wtrm.js

        while (argv.length > 0) {
            const key = argv.shift(0).replace(/^--/, "");
            if (this.config[key] === undefined) {
                this.usage(`ERR: invalid parameter: --${key}`);
            }
            if (key == "screenshots"
                    || key == "pretty"
                    || key == "dump-html-on-failed-step"
                    || key == "no-sandbox") {
                this.config[key] = true;
            } else if (argv.length == 0) {
                this.usage(`ERR: missing value for option --${key}`);
            } else {
                this.config[key] = argv.shift(0);
            }
        }

        if (this.config["workflow"] === false) {
            this.usage(`ERR: missing mandatory parameter --workflow`);
        }

        if (this.config["workflow"] === "stdin") {
            const data = []
            process.stdin.on("data", (line) => data.push(line.toString()));
            process.stdin.on("end", () => { this.steps = JSON.parse(data) });
        } else {
            const data = FS.readFileSync(this.config["workflow"], "utf8");
            this.steps = JSON.parse(data);
        }

        this.numOpenPages = 0;
        this.debug = {console:[], pageerror: [], error: [], html: []};
        this.runtime = 0; // summary of the runtime of all steps
        this.actions = new Actions();
    }

    usage(err) {
        if (err) {
            console.log(err);
        }
        console.log("\nUsage: node bloonix-wtrm [ OPTIONS ]\n")
        console.log("Options:\n")
        console.log("--page-width <pixel>");
        console.log(`  Default: ${this.config["page-width"]}\n`)
        console.log("--page-height <pixel>");
        console.log(`  Default: ${this.config["page-height"]}\n`)
        console.log("--timeout <milliseconds>");
        console.log(`  Default: ${this.config["timeout"]}\n`)
        console.log("--screenshots");
        console.log("  Flag: add a screenshot to each step.\n")
        console.log("--pretty");
        console.log("  Flag: pretty output.\n")
        console.log("--workflow <file|stdin>");
        console.log("  Pass a filename to read the workflow from.")
        console.log("  You can pass 'stdin' if you want to pass")
        console.log("  the workflow to STDIN.\n")
        console.log("--html-file <file>");
        console.log("  Pass a html-file to read the content and")
        console.log("  process the workflow on.\n")
        console.log("--host-id <id>");
        console.log(`  Default: ${this.config["host-id"]}\n`)
        console.log("--service-id <id>");
        console.log(`  Default: ${this.config["service-id"]}\n`)
        console.log("--log-file <file>");
        console.log(`  Default: ${this.config["log-file"]}\n`);
        console.log("--user-data-dir <dir>");
        console.log("  Will be passed to puppeteer.launch().")
        console.log("  Not set by default.\n")
        console.log("--no-sandbox");
        console.log("  Flag: Browser will startet with no sandbox.")
        console.log("  Not set by default.\n")
        process.exit();
    }

    async run() {
        const globalTimeout = setTimeout(() => {
            this.log({"status": "err", "message": "workflow timed out"});
            this.writeLog("workflow timed out");
            this.quitBrowser(1000);
        }, this.config["timeout"]);
        const launchArgs = {
            headless: "new",
            sandbox: true,
            args: [
                '--disable-crash-reporter',
                '--disable-breakpad',
                '--disable-dev-shm-usage',
            ]
        };
        if (this.config["no-sandbox"] === true) {
            launchArgs.args.push('--no-sandbox');
            launchArgs.sandbox = false;
        }
        if (this.config["user-data-dir"] !== false) {
            launchArgs.userDataDir = this.config["user-data-dir"];
        }
        this.browser = await PUPPETEER.launch(launchArgs);
        this.page = this.mainPage = await this.browser.newPage();
        const client = await this.page.target().createCDPSession()
        if (this.config["allow-download"] === false) {
            this.writeLog("allow-download = false")
            await client.send("Page.setDownloadBehavior", {
                behavior: "deny",
                downloadPath: "/dev/null",
            });
        } else {
            this.writeLog("allow-download = "+ this.config["allow-download"])
            await client.send("Page.setDownloadBehavior", {
                behavior: "allow",
                downloadPath: this.config["allow-download"],
            });
        }
        this.page.setDefaultNavigationTimeout(0);
        this.page.on("console", msg => this.debug.console.push(msg.text()));
        this.page.on("pageerror", msg => this.debug.pageerror.push(msg.toString()));
        this.page.on("error", msg => this.debug.error.push(msg.toString()));
        await this.page.setViewport({
            width: this.config["page-width"],
            height: this.config["page-height"]
        });
        if (this.config["html-file"]) {
            await this.page.goto(`file://${this.config["html-file"]}`);
        }
        this.writeLog("browser started");
        await this.runSteps();
        this.log({
            "status": "done",
            "debug": this.debug,
            "runtime": this.runtime
        });
        clearTimeout(globalTimeout);
        this.writeLog("stop browser");
        // process.exit terminates the Node process immediately,
        // without waiting for anything else to be written to
        // STDOUT. This can cause problems when large images
        // are written to STDOUT.
        if (this.config["screenshots"] === true) {
            this.quitBrowser(3000);
        } else {
            this.quitBrowser(500);
        }
    }

    // Unfortunately, in some transactions, after calling
    // "await browser.close()", Puppeteer throws an error,
    // as example "Error: Navigating frame was detached".
    // These errors cannot be catched with try{} and after
    // that it may happen that the browser processes not
    // actually being terminated, detaching themselves
    // and then continuing to run under PPID 1. To avoid
    // this, the main process (browser.process().pid) is
    // killed. Normally the kill command should only be
    // called in extreme situations, but due to these
    // "bugs" the browser process must always be terminated
    // with kill to avoid the risk of zombie processes
    // being created and the server running OOM.
    quitBrowser(timeout) {
        if (this.browser && this.browser.process()) {
            setTimeout(() => {
                try {
                    process.kill(this.browser.process().pid);
                } finally {
                    process.exit();
                }
            }, timeout);
        } else {
            process.exit();
        }
    }

    async runSteps() {
        this.writeLog("run steps");
        for (let i = 0; i < this.steps.length; i++) {
            const step = this.steps[i];
            let result;
            // 3 tries
            for (let t = 1; t <= 3; t++) {
                this.writeLog(`run step ${step.action} t=${t}`);
                result = await this.runStep(step);
                if (step.action == "doWaitForElement"
                        || step.action == "doSwitchToNewPage"
                        || step.action == "doSwitchToMainPage"
                        || step.action == "doSleep"
                        || result.success === true) {
                    break;
                }
                if (t < 3) {
                    await SLEEP(200);
                }
            }
            this.log({
                "status": "ok",
                "data": result
            });
            if (result.success === false) {
                if (this.config["dump-html-on-failed-step"] === true) {
                    this.debug.html.push(await this.actions.doDump(this.page));
                }
                break;
            }
            // Some applications have DoS protection that kicks
            // in when there are many requests per second.
            // The same can happen with events that are triggered
            // when form fields are filled in and make requests
            // to the web server. To prevent this, a small delay
            // between the steps is required.
            if (step.action == "doClick"
                    || step.action == "doSubmit"
                    || step.action == "doTriggerEvent"
                    || step.action == "doUrl") {
                await SLEEP(200);
            } else if (step.action == "doFill"
                    || step.action == "doCheck"
                    || step.action == "doUncheck"
                    || step.action == "doSelect") {
                await SLEEP(50);
            } else {
                await SLEEP(50);
                //await SLEEP(20);
            }
        }
        this.writeLog("steps finished");
    }

    async runStep(step) {
        if (!step.action in this.actions) {
            throw(`action ${step.action} does not exist`);
        }
        const result = {
            step: Object.assign({}, step),
            start: new Date().getTime(),
            stop: false,
            took: false,
            success: false,
            debug: [],
        };
        if (step.hideValue === true) {
            result.step.value = "xxx";
        }
        let waitForNavigation = this.page.waitForNavigation();
        try {
            let ret;
            if (step.action == "doSwitchToNewPage"
                    || step.action == "doSwitchToMainPage"
                    || step.action == "doUrl"
                    || step.action == "checkUrl"
                    || step.action == "checkDownload") {
                ret = await this.actions[step.action](step, this)
            } else {
                ret = await this.actions[step.action](step, this.page);
            }
            if (typeof ret == "object") {
                for (let a in ret) {
                    result[a] = ret[a];
                }
            } else if (typeof ret == "boolean") {
                result.success = ret;
            }
        } catch (error) {
            if (/Execution context was destroyed/.test(error.message)) {
                await waitForNavigation;
                await SLEEP(200);
                return await this.runStep(step);
            }
            result.message = error.toString();
        };
        result.stop = new Date().getTime();
        result.took = result.stop - result.start;
        this.runtime += result.took;
        if (this.config["screenshots"] === true) {
            result.image = await this.page.screenshot({encoding: "base64"});
        }
        if (step.action == "doUrl") {
            await this.updateNumberOfOpenPages();
        }
        return result;
    }

    async updateNumberOfOpenPages() {
        const pages = this.numOpenPages = await this.browser.pages();
        this.numOpenPages = pages.length;
        if (this.numOpenPages > 10) {
            throw("the maximum number of open pages exceeded");
        }
    }

    log(data) {
        const indent = this.config["pretty"] === true ? 2 : null;
        console.log(JSON.stringify(data, null, indent));
    }

    writeLog(message) {
        const pid = this.getBrowserPid();
        const hostId = this.config["host-id"];
        const serviceId = this.config["service-id"];
        const time = this.getTimestamp();
        const msg = `${time} p=${pid} h=${hostId} s=${serviceId} ${message}\n`;
        if (this.config["log-file"] == "stdout") {
            FS.writeSync(1, msg);
        } else if (this.config["log-file"] != "none") {
            FS.writeFileSync(this.config["log-file"], msg, {flag: "a+"}, err => {});
        }
    }

    getBrowserPid() {
        if (this.browser && this.browser.process()) {
            return this.browser.process().pid;
        }
        return 0;
    }

    getTimestamp() {
        const now = new Date();
        const year = now.getYear() + 1900;
        const month = this.zeroPad(now.getMonth() + 1);
        const day = this.zeroPad(now.getDate());
        const hours = this.zeroPad(now.getHours());
        const minutes = this.zeroPad(now.getMinutes());
        const seconds = this.zeroPad(now.getSeconds());
        return `${year}-${month}-${day} ${hours}:${minutes}:${seconds}`;
    }

    zeroPad(num) {
        if (num < 10) {
            return "0" + num;
        }
        return num;
    }
}

class Actions {
    constructor() {
        const fns = [
            "doFill",
            "doClick",
            "doSubmit",
            "doCheck",
            "doUncheck",
            "doSelect",
            "doTriggerEvent",
            "checkIfElementExists",
            "checkIfElementNotExists",
            "checkIfElementHasText",
            "checkIfElementMatchText",
            "checkIfElementHasNotText",
            "checkIfElementHasHTML",
            "checkIfElementHasNotHTML",
            "checkIfElementHasValue",
            "checkIfElementHasNotValue",
            "checkIfElementIsChecked",
            "checkIfElementIsNotChecked",
            "checkIfElementIsSelected",
            "checkIfElementIsNotSelected",
        ];

        for (let i = 0; i < fns.length; i++) {
            const name = fns[i];
            this[name] = this.evaluate;
        }
    }

    async doAuth(step, page) {
        await page.authenticate({username: step.username, password: step.password});
        return true;
    }

    async doUserAgent(step, page) {
        await page.setUserAgent(step.userAgent);
        return true;
    }

    async doSleep(step, page) {
        await new Promise(resolve => setTimeout(resolve, step.ms));
        return true;
    }

    async doWaitForElement(step, page) {
        const timeout = step.timeout === undefined
            ? 30000
            : step.timeout;

        const ms = step.sleep === undefined
            ? 100
            : step.sleep;

        const expire = new Date().getTime() + timeout;
        let ret = false;

        while (new Date().getTime() < expire) {
            ret = await this.evaluate(step, page);
            if (ret && ret.success === true) {
                break;
            }
            await SLEEP(ms);
        }

        return ret;
    }

    async doSwitchToFrame(step, page) {
        return {success: false, message: "this feature is not supported"};
    }

    async doSwitchToParentFrame(step, page) {
        return {success: false, message: "this feature is not supported"};
    }

    async doSwitchToNewPage(step, parent) {
        const expire = new Date().getTime() + 15000;

        while (new Date().getTime() < expire) {
            const openPages = [], pages = await parent.browser.pages();

            for (let page of pages) {
                openPages.push(page);

                // if step.url is set
                if (step.url == page.url()) {
                    parent.page = page;
                    return {success: true, message: "switch to "+ parent.page.url()}
                }
            }

            if (!step.url) {
                if (openPages.length > parent.numOpenPages) {
                    parent.page = openPages[openPages.length - 1];
                    return {success: true, message: "switch to "+ parent.page.url()}
                }
            }

            await SLEEP(100);
        }

        return {success: false};
    }

    async doSwitchToMainPage(step, parent) {
        parent.page = parent.mainPage;
        return {success: true, message: "switch back to "+ parent.page.url(), debug: []}
    }

    async doAddCookie(step, page) {
        await page.setCookie({
            name: step.name,
            value: step.value,
            domain: step.domain,
            path: "/",
            expires: (new Date()).getTime() + (1000 * 60 * 5) // expires in 5 minutes,
        });
        return true;
    }

    async checkUrl(step, parent) {
        const page = parent.page;
        const res = parent.lastGotoResult;
        const header = res.headers();
        let success = true;

        const href = await page.evaluate(() => {
            return document.location.href;
        });

        if (step.status && parseInt(step.status) !== res.status()) {
            success = false;
        }
        if (step.contentType && step.contentType !== header["content-type"]) {
            success = false;
        }
        if (href !== step.url) {
            success = false;
        }

        return {success: success, message: "url: "+ href};
    }

    async doUrl(step, parent) {
        const page = parent.page;
        const res = await page.goto(step.url);
        const header = res.headers();
        const debug = ["status: "+ res.status()];
        parent.lastGotoResult = res;
        let success = true;
        for (let key in header) {
            debug.push(key + ": "+ header[key]);
        }
        if (step.status && parseInt(step.status) !== res.status()) {
            success = false;
        }
        if (step.contentType && step.contentType !== header["content-type"]) {
            success = false;
        }
        return {success: success, debug: debug};
    }

    async checkDownload(step, parent) {
        if (parent.config["allow-download"] === false) {
            return {success: false, debug: ["download disabled"]};
        }
        if (!step.filename) {
            return {success: false, debug: ["missing file name"]};
        }
        const file = parent.config["allow-download"] + "/" + step.filename;
        if (FS.existsSync(file)) {
            return {success: true, debug: ["file exists"]}
        }
        return {success: true, debug: ["file does not exists"]}
    }

    // only for internal usage
    async doDump(page) {
        return await page.evaluate(() => document.body.innerHTML);
    }

    // only for internal usage
    async doScreenshot(step, page) {
        await page.screenshot({path: "example.png"});
        return true;
    }

    async evaluate(step, page) {
        if (step.action === "doFill" && !step.parent && !/^</.test(step.element)) {
            let sel;
            if (/^[\[\.#]/.test(step.element)) {
                sel = step.element;
            } else {
                sel = "[name='"+ step.element +"']";
            }
            await page.type(sel, step.value);
            return {success: true};
        }

        return await page.evaluate((step) => {
            let element = false,
                elements = [],
                tagName = false,
                foundById = false,
                result = {success: false, debug: []};

            const _document = step.parent
                ? document.getElementById(step.parent.replace(/^#/, ""))
                : document;

            if (!_document) {
                result.message = "parent element does not exist";
                return result;
            }

            /*    Valid:
                    <h1>
                    <h1[1]>     max 999
                    <h1[1] class="foo">
                    <h1 id="abc" class="foo">

                Invalid:
                    <h1 id='abc" class="foo">
                    <h1 id="abc" x s s="">
                    <h1 id="abc'>
            */

            const tagRx = /^<\s*([a-zA-Z0-9]+)\s*>/,
                tagNumRx = /^<\s*([a-zA-Z0-9]+)\[(\d{1,3})\]\s*>/,
                tagAttrRx = /^<\s*([a-zA-Z0-9]+)(?:\[(\d{1,3})\]){0,1}\s+(.+)\s*>/;

            if (step.element) {
                if (/^#/.test(step.element)) {
                    element = _document.getElementById(step.element.replace(/^#/, ""));
                    elements = [element];
                    foundById = true;
                } else if (/^\./.test(step.element)) {
                    elements = _document.getElementsByClassName(step.element.replace(/^\./, ""));
                    element = elements[0];
                } else if (/^[a-zA-Z0-9_\-]+/.test(step.element)) {
                    if (step.parent) {
                        elements = _document.querySelectorAll(
                            `input[name='${step.element}'], textarea[name='${step.element}']`,
                        );
                        element = elements[0];
                    } else {
                        elements = _document.getElementsByName(step.element);
                        element = elements[0];
                    }
                } else if (tagRx.test(step.element)) {
                    const match = tagRx.exec(step.element),
                        tag = match[1];
                    elements = _document.getElementsByTagName(tag);
                    element = elements[0];
                } else if (tagNumRx.test(step.element)) {
                    const match = tagNumRx.exec(step.element),
                        tag = match[1],
                        num = match[2];
                    elements = _document.getElementsByTagName(tag);
                    element = elements[num];
                } else if (tagAttrRx.test(step.element)) {
                    const match = tagAttrRx.exec(step.element),
                        tag = match[1],
                        num = match[2],
                        attrs = match[3];

                    const tags = _document.getElementsByTagName(tag);

                    if (tags.length > 0) {
                        const a = attrs.match(/([a-zA-Z0-9_\-]+=(?:'[^']*'|"[^"]*"))/g),
                            hasAttrs = [];

                        for (let i = 0; i < a.length; i++) {
                            a[i] = a[i].replace(/^\s+/, "");
                            a[i] = a[i].replace(/\s+$/, "");
                            const b = a[i].match(/([a-zA-Z0-9_\-]+)=(.+)/);
                            const key = b[1],
                                value = b[2].replace(/["']/g, "");
                            hasAttrs.push({attr: key, value: value});
                        }

                        for (let x = 0; x < tags.length; x++) {
                            const obj = tags[x];
                            let foundAttrs = 0;

                            for (let y in hasAttrs) {
                                const attr = hasAttrs[y];
                                if (obj.getAttribute(attr.attr) === attr.value) {
                                    foundAttrs += 1;
                                }
                            }

                            if (foundAttrs === hasAttrs.length) {
                                elements.push(obj);
                            }
                        }
                    }

                    if (elements.length > 0) {
                        if (num) {
                            element = elements[num - 1];
                        } else {
                            element = elements[0];
                        }
                    }
                }
            }

            if (!element) {
                if (step.action != "checkIfElementExists" &&
                        step.action != "checkIfElementNotExists" &&
                        step.action != "doWaitForElement") {
                    result.message = "element does not exist";
                    return result;
                }
            } else {
                tagName = element.tagName.toLowerCase();
                result.debug.push(tagName +" tag found");
            }

            // Action: doFill
            if (step.action === "doFill") {
                if (tagName != "input" && tagName != "textarea") {
                    result.message = "tag '"+ tagName +"' is not a valid form element";
                } else {
                    element.value = step.value;
                    result.success = true;
                }
            }

            // Action: doClick
            else if (step.action === "doClick") {
                result.message = "unable to click element";
                element.addEventListener("click", () => {
                    result.message = "click was successful";
                    result.success = true;
                });
                element.click();
            }

            // Action: doSubmit
            else if (step.action === "doSubmit") {
                element.submit();
                result.success = true;
            }

            // Action: doCheck, doUncheck
            else if (step.action === "doCheck" || step.action == "doUncheck") {
                if (foundById === true) {
                    if (element.value == step.value) {
                        element.checked = step.action === "doCheck" ? true : false;
                        result.success = true;
                    }
                } else {
                    for (let i = 0; i < elements.length; i++) {
                        const obj = elements[i];
                        if (obj.value == step.value) {
                            obj.checked = step.action == "doCheck" ? true : false;
                            result.success = true;
                        }
                    }
                }
            }

            // Action: doSelect
            else if (step.action === "doSelect") {
                if (tagName != "select") {
                    result.message = "tag '"+ tagName +"' is not a valid form element";
                } else {
                    for (let i = 0; i < element.options.length; i++) {
                        if (element.options[i].value == step.value) {
                            element.selectedIndex = i;
                            result.success = true;
                        }
                    }
                }
            }

            // Action: doTriggerEvent
            else if (step.action == "doTriggerEvent") {
                if (step["event"] == "focus") {
                    step.focus();
                } else {
                    const ev = document.createEvent("UIEvents");
                    ev.initUIEvent(step["event"], true, true, window, 1);
                    element.dispatchEvent(event);
                }
                result.success = true;
            }

            // Action: checkIfElementExists
            else if (step.action == "checkIfElementExists") {
                if (element) {
                    result.success = true;
                }
            }

            // Action: checkIfElementExists, doWaitForElement
            else if (step.action == "doWaitForElement") {
                if (element) {
                    if (step.text !== undefined && step.text.length) {
                        for (let i = 0; i < elements.length; i++) {
                            //console.log(
                            //        "DEBUG wait",
                            //        elements[i].tagName,
                            //        elements[i].innerText,
                            //        "===", step.text);
                            if (elements[i].innerText == step.text) {
                                result.success = true;
                                break;
                            }
                        }
                    } else {
                        result.success = true;
                    }
                }
            }

            // Action: checkIfElementNotExists
            else if (step.action == "checkIfElementNotExists") {
                if (!element) {
                    result.success = true;
                }
            }

            // Action: checkIfElementHasText
            else if (step.action == "checkIfElementHasText") {
                if (element.innerText == step.text) {
                    result.success = true;
                }
            }

            // Action: checkIfElementHasNotText
            else if (step.action == "checkIfElementHasNotText") {
                if (element.innerText != step.text) {
                    result.success = true;
                }
            }

            // Action: checkIfElementMatchText
            else if (step.action == "checkIfElementMatchText") {
                const re = new RegExp(step.text, "g");
                if (element.innerText.match(re)) {
                    result.success = true;
                }
            }

            // Action: checkIfElementHasElement
            else if (step.action == "checkIfElementHasHTML") {
                const html = element.innerHTML.replace(/(\r\n|\n|\r)$/, "");
                if (html == step.html) {
                    result.success = true;
                }
            }

            // Action: checkIfElementHasNotElement
            else if (step.action == "checkIfElementHasNotHTML") {
                const html = element.innerHTML.replace(/(\r\n|\n|\r)$/, "");
                if (html != step.html) {
                    result.success = true;
                }
            }

            // Action: checkIfElementHasValue
            else if (step.action == "checkIfElementHasValue") {
                if (tagName != "input" && tagName != "textarea") {
                    result.message = "tag '"+ tagName +"' is not a valid form element";
                } else {
                    if (element.value == step.value) {
                        result.success = true;
                    }
                }
            }

            // Action: checkIfElementHasNotValue
            else if (step.action == "checkIfElementHasNotValue") {
                if (tagName != "input" && tagName != "textarea") {
                    result.message = "tag '"+ tagName +"' is not a valid form element";
                } else {
                    if (element.value != step.value) {
                        result.success = true;
                    }
                }
            }

            // Action: checkIfElementIsChecked
            else if (step.action == "checkIfElementIsChecked") {
                if (foundById === true) {
                    if (element.checked === true) {
                        result.success = true;
                    }
                } else {
                    for (let i = 0; i < elements.length; i++) {
                        const obj = elements[i];
                        if (obj.value == step.value) {
                            if (obj.checked) {
                                result.success = true;
                            }
                            break;
                        }
                    }
                }
            }

            // Action: checkIfElementIsNotChecked
            else if (step.action == "checkIfElementIsNotChecked") {
                if (foundById === true) {
                    if (element.checked !== true) {
                        result.success = true;
                    }
                } else {
                    for (let i = 0; i < elements.length; i++) {
                        const obj = elements[i];
                        if (obj.value == step.value) {
                            if (!obj.checked) {
                                result.success = true;
                            }
                            break;
                        }
                    }
                }
            }

            // Action: checkIfElementIsSelected
            else if (step.action == "checkIfElementIsSelected") {
                if (tagName != "select") {
                    result.message = "tag '"+ tagName +"' is not a valid form element";
                } else if (element.options[element.selectedIndex].value == step.value) {
                    result.success = true;
                }
            }

            // Action: checkIfElementIsNotSelected
            else if (step.action == "checkIfElementIsNotSelected") {
                if (tagName != "select") {
                    result.message = "tag '"+ tagName +"' is not a valid form element";
                } else if (element.options[element.selectedIndex].value != step.value) {
                    result.success = true;
                }
            }

            return result;
        }, step);
    }
};

const wtrm = new WTRM();
wtrm.run();
